﻿using Barsalt.WindowSystem.Components;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using UnityEngine;

namespace Barsalt.WindowSystem {
	/// <summary>
	/// The BaseWindow provides a class for ingame windows in Unity3D. It provides several different tools. The IdSystem that provides each window with a unique ID. The ResizeSystem that allows the window to be resized by mouse action. 
	/// </summary>
	public abstract class BaseWindow {
		#region CLASS_ATTRIBUTES
		//==========================================================================
		private static readonly IdSystem ID_SYSTEM			= new IdSystem(100);
		private static readonly FocusSystem FOCUS_SYSTEM	= new FocusSystem();
		private static readonly int MENU_BUTTON_WIDTH		= 16;
		private static readonly int MENU_BUTTON_HEIGHT		= 16;
		private static readonly int MENU_BUTTON_PADDING		= 4;
		//==========================================================================
		#endregion

		#region OBJECT_ATTRIBUTES
		//==========================================================================
		private bool isEnabled			= true;
		private bool isFocused			= false;
		private int id;
		private Rect dimension;
		private Rect exitButton;
		private string title;
		private bool isDragable;
		private bool isResizeable;
		private IdSystem idSystem;
		private bool showExitButton		= true;
		private Vector2 minimalSize		= Vector2.zero;
		private Vector2 maximalSize		= Vector2.zero;
		private ResizeSystem resizer	= new ResizeSystem();
		//==========================================================================
		#endregion

		#region CONSTRUCTOR
		//==========================================================================
		/// <summary>
		/// 
		/// </summary>
		/// <param name="dimension">The position and size of the window</param>
		/// <param name="title">The title of the window</param>
		/// <param name="id">The ID for the window</param>
		public BaseWindow(Rect dimension, string title, int id) : this(dimension, title) {
			this.idSystem	= ID_SYSTEM;
			this.id			= id;

			idSystem.AddClaimID_Handler(IDChangeCallback);

			this.Dimension	= dimension;

			this.title		= title;

			FOCUS_SYSTEM.Add(this);
		}

		/// <summary>
		/// 
		/// </summary>
		/// <param name="dimension">The position and size of the window</param>
		/// <param name="title">The title of the window</param>
		/// <param name="manager">The ID Manager that is to be used by the window</param>
		public BaseWindow(Rect dimension, string title, IdSystem manager) : this(dimension, title) {
			this.idSystem	= manager;
			this.id			= idSystem.GetID();

			idSystem.AddClaimID_Handler(IDChangeCallback);

			this.Dimension	= dimension;
			this.title		= title;
		}

		/// <summary>
		/// 
		/// </summary>
		/// <param name="dimension">The position and size of the window</param>
		/// <param name="title">The title of the window</param>
		public BaseWindow(Rect dimension, string title) {
			this.idSystem	= ID_SYSTEM;
			this.id			= idSystem.GetID();

			idSystem.AddClaimID_Handler(IDChangeCallback);

			this.Dimension	= dimension;
			this.title		= title;
		}
		//==========================================================================
		#endregion

		#region OBJECT_FUNCTIONS
		//==========================================================================
		/// <summary>
		/// The function displays the contents of the window if the window is enabled. The function calls the OnGui function to allow user implemented content.
		/// </summary>
		public void Display() {
			if(IsEnabled)
				Dimension = GUI.Window(id, dimension, DisplayConfiguration, title);
		}

		private void DisplayConfiguration(int id){
			// check for mouse action and set focus 
			if(Event.current.isMouse && Event.current.type == EventType.MouseDown ) {
				FOCUS_SYSTEM.GetFocus(this);
				GUI.FocusWindow(ID);
			}
		
			OnGui(id);

			if(showExitButton)
				ExitButton();

			if(isDragable && !resizer.IsResizing)
				GUI.DragWindow();

			if(isResizeable && isFocused)
				Dimension = resizer.Resize(Dimension);
		}

		/// <summary>
		/// The function contains the user implemented content and is called by the window after the call of the Display() function. 
		/// </summary>
		/// <param name="id"></param>
		protected abstract void OnGui(int id);

		/// <summary>
		/// The function returns the ID to the IdSystem of destruction.
		/// </summary>
		public void finalize(){
			idSystem.ReturnID(this.ID);

			FOCUS_SYSTEM.Remove(this);
		}

		private void ExitButton(){
			if(GUI.Button(exitButton, "-"))
				IsEnabled = false;
		}

		protected virtual void OnDimensionChange(){}
		//==========================================================================
		#endregion

		#region GETTER_AND_SETTER
		//==========================================================================
		public int ID{
			get{
				return id;
			}
			set{
				id = value;

				idSystem.ClaimID(id);
			}
		}

		public Rect Dimension{
			get{
				return dimension;
			}
			set{
				if(!dimension.Equals(value)){
					if(!minimalSize.Equals(Vector2.zero) && value.width < minimalSize.x){
						value.width		= minimalSize.x;
						value.x			= dimension.x;
					}

					if(!minimalSize.Equals(Vector2.zero) && value.height < minimalSize.y) {
						value.height	= minimalSize.y;
						value.y			= dimension.y;
					}
					
					if(!maximalSize.Equals(Vector2.zero) && value.width > maximalSize.x) {
						value.width		= maximalSize.x;
						value.x			= dimension.x;
					}

					if(!maximalSize.Equals(Vector2.zero) && value.height > maximalSize.y) {
						value.height	= maximalSize.y;
						value.y			= dimension.y;
					}

					if(!dimension.Equals(value)){
						dimension		= value;
						exitButton		= new Rect(dimension.width - MENU_BUTTON_WIDTH - MENU_BUTTON_PADDING, MENU_BUTTON_PADDING, MENU_BUTTON_WIDTH, MENU_BUTTON_HEIGHT);

						OnDimensionChange();
					}
				}
			}
		}

		public string Title{
			get{
				return title;
			}
			set{
				title = value;
			}
		}

		public bool IsDragAble{
			get{
				return isDragable;
			}
			set{
				isDragable = value;
			}
		}

		public bool IsResizeAble {
			get {
				return isResizeable;
			}
			set {
				isResizeable = value;
			}
		}

		public bool IsEnabled{
			get{
				return isEnabled;
			}
			set{
				isEnabled =  value;
			}
		}

		public bool IsFocused{
			get{
				return isFocused;
			}
			set{
				isFocused = value;
			}
		}

		public void IDChangeCallback(int id){
			if(this.id == id){
				id = idSystem.GetID();
			}
		}

		public bool ExitButtonEnabled{
			get{
				return showExitButton;
			}
			set{
				showExitButton = value;
			}
		}

		public Vector2 MinimalSize{
			get{
				return minimalSize;
			}
			set{
				if(value.x >= 0 && value.y >= 0 ){
					minimalSize = value;
					
					if(!maximalSize.Equals(Vector2.zero)){
						if(maximalSize.x < minimalSize.x){
							maximalSize = new Vector2(minimalSize.x, maximalSize.y);
						}

						if(maximalSize.y < minimalSize.y) {
							maximalSize = new Vector2(maximalSize.x, minimalSize.y);
						}
					}

					Rect resizedDimension = new Rect(dimension);

					if(dimension.width < minimalSize.x){
						resizedDimension = new Rect(resizedDimension.x, resizedDimension.y, minimalSize.x, resizedDimension.height);
					}

					if(dimension.height < minimalSize.y) {
						resizedDimension = new Rect(resizedDimension.x, resizedDimension.y, resizedDimension.width, minimalSize.y);
					}

					Dimension = resizedDimension;
				}
			}
		}

		public Vector2 MaximalSize{
			get{
				return maximalSize;
			}
			set{
				if(value.x >= 0 && value.y >= 0){
					maximalSize = value;

					if(!minimalSize.Equals(Vector2.zero)){
						if(minimalSize.x > maximalSize.x) {
							minimalSize = new Vector2(maximalSize.x, minimalSize.y);
						}

						if(minimalSize.y > maximalSize.y) {
							minimalSize = new Vector2(minimalSize.x, maximalSize.y);
						}
					}

					Rect resizedDimension = new Rect(dimension);

					if(dimension.width > maximalSize.x) {
						resizedDimension = new Rect(resizedDimension.x, resizedDimension.y, maximalSize.x, resizedDimension.height);
					}

					if(dimension.height > maximalSize.y) {
						resizedDimension = new Rect(resizedDimension.x, resizedDimension.y, resizedDimension.width, maximalSize.y);
					}

					Dimension = resizedDimension;
				}
			}
		}

		public int ResizeHandleSize{
			get{
				return resizer.HandleSize;
			}
			set{
				resizer.HandleSize = value;
			}
		}
		//==========================================================================
		#endregion
	}
}
