﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using UnityEngine;

namespace Barsalt.WindowSystem.Components {
	/// <summary>
	/// The ResizeSystem resizes a Rect based on the position and activates buttons of the mouse. The interface for resizeing is the Resize() function.  
	/// </summary>
	public sealed class ResizeSystem {
		#region RESIZE_BEHAVIOR_CLASSES
		//==========================================================================
		/// <summary>
		/// The ResizeBehavior is the base class for all classes that manage a resize action. It exists a class for each possible direction.
		/// </summary>
		private abstract class ResizeBehavior{
			private Vector2 lastPosition = Vector2.zero;

			/// <summary>
			/// The regulation of the actual resize process.
			/// </summary>
			/// <param name="dimension">The original Rect</param>
			/// <param name="translatedPosition">The position of the mouse in the original Rect</param>
			/// <returns>The resized Rect</returns>
			public abstract Rect Resize(Rect dimension, Vector2 translatedPosition);

			/// <summary>
			/// The function check if the position is contained in its activation area.
			/// </summary>
			/// <param name="Dimension">The original Rect</param>
			/// <param name="translatedPosition">The position of the mouse in the original Rect</param>
			/// <param name="handleSize">The size of the activation area</param>
			/// <returns></returns>
			public abstract bool Contains(Rect Dimension, Vector2 translatedPosition, int handleSize);

			/// <summary>
			/// The last position of the mouse in the Rect
			/// </summary>
			public Vector2 LastPosition{
				get{
					return lastPosition;
				}
				set{
					lastPosition = value;
				}
			}
		}

		private class North : ResizeBehavior{
			public override Rect Resize(Rect dimension, Vector2 translatedPosition) {
				Vector2 differnce = translatedPosition - LastPosition;

				LastPosition = translatedPosition;

				if(differnce.y != 0) {
					if((translatedPosition.y < 0 && differnce.y < 0) || differnce.y > 0) {
						LastPosition = new Vector2(LastPosition.x, LastPosition.y - differnce.y);

						return new Rect(dimension.x, dimension.y + differnce.y, dimension.width, dimension.height - differnce.y);
					}
				}

				return dimension;
			}

			public override bool Contains(Rect dimension, Vector2 translatedPosition, int handleSize) {
				if(	translatedPosition.x > handleSize && translatedPosition.x < dimension.width - 2 * handleSize &&
					translatedPosition.y > 0 && translatedPosition.y < handleSize){
					return true;	
				} else {
					return false;
				}
			}
		}

		private class NorthEast : ResizeBehavior {
			public override Rect Resize(Rect dimension, Vector2 translatedPosition) {
				Vector2 differnce = translatedPosition - LastPosition;

				LastPosition = translatedPosition;

				if(differnce.x != 0 || differnce.y != 0) {
					float y = dimension.y;
					float width = dimension.width;
					float height = dimension.height;

					if((translatedPosition.x > dimension.width && differnce.x > 0) || differnce.x < 0) {
						width += differnce.x;
					}

					if((translatedPosition.y < 0 && differnce.y < 0) || differnce.y > 0) {
						LastPosition = new Vector2(LastPosition.x, LastPosition.y - differnce.y);

						y += differnce.y;
						height -= differnce.y;
					}
					
					return new Rect(dimension.x, y, width, height);
				}

				return dimension;
			}

			public override bool Contains(Rect dimension, Vector2 translatedPosition, int handleSize) {
				if(translatedPosition.x > dimension.width - handleSize && translatedPosition.x < dimension.width &&
					translatedPosition.y > 0 && translatedPosition.y < handleSize) {
					return true;
				} else {
					return false;
				}
			}
		}

		private class East : ResizeBehavior {
			public override Rect Resize(Rect dimension, Vector2 translatedPosition) {
				Vector2 differnce = translatedPosition - LastPosition;

				LastPosition = translatedPosition;

				if(differnce.x != 0) {
					if((translatedPosition.x > dimension.width && differnce.x > 0) || differnce.x < 0) {
						return new Rect(dimension.x, dimension.y, dimension.width + differnce.x, dimension.height);
					}
				}

				return dimension;
			}

			public override bool Contains(Rect dimension, Vector2 translatedPosition, int handleSize) {
				if(translatedPosition.x > dimension.width - handleSize && translatedPosition.x < dimension.width &&
					translatedPosition.y > handleSize && translatedPosition.y < dimension.height - handleSize) {
					return true;
				} else {
					return false;
				}
			}
		}

		private class SouthEast : ResizeBehavior {
			public override Rect Resize(Rect dimension, Vector2 translatedPosition) {
				Vector2 differnce = translatedPosition - LastPosition;

				LastPosition = translatedPosition;

				if(differnce.x != 0 || differnce.y != 0) {
					float width = dimension.width;
					float height = dimension.height;

					if((translatedPosition.x > dimension.width && differnce.x > 0) || differnce.x < 0) {
						width += differnce.x;
					}

					if((translatedPosition.y > dimension.height && differnce.y > 0) || differnce.y < 0) {
						height += differnce.y;
					}

					return new Rect(dimension.x, dimension.y, width, height);
				}

				return dimension;
			}

			public override bool Contains(Rect dimension, Vector2 translatedPosition, int handleSize) {
				if(translatedPosition.x > dimension.width - handleSize && translatedPosition.x < dimension.width &&
					translatedPosition.y >dimension.height - handleSize && translatedPosition.y < dimension.height) {
					return true;
				} else {
					return false;
				}
			}
		}

		private class South : ResizeBehavior {
			public override Rect Resize(Rect dimension, Vector2 translatedPosition) {
				Vector2 differnce = translatedPosition - LastPosition;

				LastPosition = translatedPosition;

				if(differnce.y != 0) {
					if((translatedPosition.y > dimension.height && differnce.y > 0) || differnce.y < 0) {
						return new Rect(dimension.x, dimension.y, dimension.width, dimension.height + differnce.y);
					}
				}

				

				return dimension;
			}

			public override bool Contains(Rect dimension, Vector2 translatedPosition, int handleSize) {
				if(translatedPosition.x > handleSize && translatedPosition.x < dimension.width - handleSize &&
					translatedPosition.y > dimension.height - handleSize && translatedPosition.y < dimension.height) {
					return true;
				} else {
					return false;
				}
			}
		}

		private class SouthWest : ResizeBehavior {
			public override Rect Resize(Rect dimension, Vector2 translatedPosition) {
				Vector2 differnce = translatedPosition - LastPosition;

				LastPosition = translatedPosition;

				if(differnce.x != 0 || differnce.y != 0) {
					float x = dimension.x;
					float width = dimension.width;
					float height = dimension.height;

					if((translatedPosition.y > dimension.height && differnce.y > 0) || differnce.y < 0) {
						height += differnce.y;
					}

					if((translatedPosition.x < 0 && differnce.x < 0) || differnce.x > 0) {
						LastPosition = new Vector2(LastPosition.x - differnce.x, LastPosition.y);

						x += differnce.x;
						width -= differnce.x;
					}

					return new Rect(x, dimension.y, width, height);
				}

				return dimension;
			}

			public override bool Contains(Rect dimension, Vector2 translatedPosition, int handleSize) {
				if(translatedPosition.x > 0 && translatedPosition.x < handleSize &&
					translatedPosition.y > dimension.height - handleSize && translatedPosition.y < dimension.height) {
					return true;
				} else {
					return false;
				}
			}
		}
		
		private class West : ResizeBehavior {
			public override Rect Resize(Rect dimension, Vector2 translatedPosition) {
				Vector2 differnce = translatedPosition - LastPosition;

				LastPosition = translatedPosition;

				if(differnce.x != 0) {
					if((translatedPosition.x < 0 && differnce.x < 0) || differnce.x > 0) {
						LastPosition = new Vector2(LastPosition.x - differnce.x, LastPosition.y);

						return new Rect(dimension.x + differnce.x, dimension.y, dimension.width - differnce.x, dimension.height);
					}
				}

				return dimension;
			}

			public override bool Contains(Rect dimension, Vector2 translatedPosition, int handleSize) {
				if(translatedPosition.x > 0 && translatedPosition.x < handleSize &&
					translatedPosition.y > handleSize && translatedPosition.y < dimension.height - handleSize) {
					return true;
				} else {
					return false;
				}
			}
		}
		
		private class NorthWest : ResizeBehavior {
			public override Rect Resize(Rect dimension, Vector2 translatedPosition) {
				Vector2 differnce = translatedPosition - LastPosition;

				LastPosition = translatedPosition;

				if(differnce.x != 0 || differnce.y != 0) {
					float y = dimension.y;
					float x = dimension.x;
					float width = dimension.width;
					float height = dimension.height;

					if((translatedPosition.y < 0 && differnce.y < 0) || differnce.y > 0) {
						LastPosition = new Vector2(LastPosition.x, LastPosition.y - differnce.y);

						y += differnce.y;
						height -= differnce.y;
					}

					if((translatedPosition.x < 0 && differnce.x < 0) || differnce.x > 0) {
						LastPosition = new Vector2(LastPosition.x - differnce.x, LastPosition.y);

						x += differnce.x;
						width -= differnce.x;
					}

					return new Rect(x, y, width, height);
				}

				return dimension;
			}

			public override bool Contains(Rect dimension, Vector2 translatedPosition, int handleSize) {
				if(translatedPosition.x > 0 && translatedPosition.x < handleSize &&
					translatedPosition.y > 0 && translatedPosition.y < handleSize) {
					return true;
				} else {
					return false;
				}
			}
		}
		//==========================================================================
		#endregion

		#region OBJECT_ATTRIBUTES
		//==========================================================================
		private int handleSize									= 4;
		private bool isResizing									= false;
		private readonly List<ResizeBehavior> RESIZE_DIRECTIONS = new List<ResizeBehavior>();
		private ResizeBehavior current							= null;
		//==========================================================================
		#endregion
	
		#region CONSTRUCTORS
		//==========================================================================
		public ResizeSystem() : this(4){}

		/// <summary>
		/// 
		/// </summary>
		/// <param name="handleSize">The depth of the activation area</param>
		public ResizeSystem(int handleSize){
			this.handleSize = handleSize;

			RESIZE_DIRECTIONS.Add(new North());
			RESIZE_DIRECTIONS.Add(new NorthEast());
			RESIZE_DIRECTIONS.Add(new East());
			RESIZE_DIRECTIONS.Add(new SouthEast());
			RESIZE_DIRECTIONS.Add(new South());
			RESIZE_DIRECTIONS.Add(new SouthWest());
			RESIZE_DIRECTIONS.Add(new West());
			RESIZE_DIRECTIONS.Add(new NorthWest());
		}
		//==========================================================================
		#endregion
	
		#region OBJECT_FUNCTIONS
		//==========================================================================
		/// <summary>
		/// The function manages the resize process by activating when the mouse is clicked and within an activation area.
		/// </summary>
		/// <param name="dimension">The original Rect that is to be resized</param>
		/// <returns>The resized Rect</returns>
		public Rect Resize(Rect dimension) {

			Vector2 translatedPosition = new Vector2((Input.mousePosition.x - dimension.x), (Screen.height - Input.mousePosition.y) - dimension.y);

			if(isResizing && current != null){
				if(!Input.GetMouseButton(0)){
					isResizing = false;
				}
				
				dimension = current.Resize(dimension, translatedPosition);
			} else if (Input.GetMouseButton(0)){
				bool isSearching = true;

				for(int index = 0; index < RESIZE_DIRECTIONS.Count && isSearching; index++){
					if(RESIZE_DIRECTIONS[index].Contains(dimension, translatedPosition, handleSize)){
						current					= RESIZE_DIRECTIONS[index];
						current.LastPosition	= translatedPosition;

						isResizing				= true;
						isSearching				= false;
					}
				}
			}

			return dimension;
		}
		//==========================================================================
		#endregion
	
		#region GETTER_AND_SETTER
		//==========================================================================
		public bool IsResizing{
			get{
				return isResizing;
			}
		}

		public int HandleSize {
			get {
				return handleSize;
			}
			set {
				handleSize = value;
			}
		}
		//==========================================================================
		#endregion	
	}
}